#ifndef FARRAY_HPP
#define FARRAY_HPP

#include <fstream.h>
#include <stdlib.h> // size_t

template <class T> class FileArrayElementAccess {};

template <class T> class FileArray;
template <class T> class FileArrayProxy
{
public:
  operator T() const;
  FileArrayProxy<T>& operator=(const T& t);
  FileArrayProxy<T>& operator=(const FileArrayProxy<T>& p);
private:
  FileArrayProxy(FileArray<T>& f, size_t i);

  operator T&() const;
  T* operator&(void) const;
  size_t index;
  FileArray<T>& fa;

  friend class FileArray<T>;
};

template <class T>
ostream& operator<<(ostream& os, const FileArrayProxy<T>& p)
{
  return os << p.operator T();
}

template <class T>
class ptr
{
public:
  ptr(T*);
  ~ptr();
  T& operator*() const;
private:
  ptr(const ptr<T>&);
  ptr<T>& operator=(const ptr<T>&);
  T* p;
};

template <class T>
ptr<T>::ptr(T* tp)
  : p(tp)
{
}

template <class T>
ptr<T>::~ptr()
{
  delete p;
}

template <class T>
T& ptr<T>::operator*() const
{
  return *p;
}

template <class T>
class FileArray
{
public:
  FileArray(const char* name, size_t size); // create one.
  FileArray(const char* name); // use existing array
  ~FileArray(void);
  T operator[](size_t size) const;
  FileArrayProxy<T> operator[](size_t size);
  size_t size() const;
private:
  FileArray(const FileArray<T>&); // illegal
  FileArray<T>& operator=(const FileArray<T>&); // illegal
  T readElement(size_t index) const;
  void storeElement(size_t index, const T& elem);
  ptr<iostream> pstream;
  size_t array_size;

  friend class FileArrayProxy<T>;
};

template <class T>
FileArray<T>::~FileArray()
{
}

template <class T>
FileArrayProxy<T>::FileArrayProxy(FileArray<T>& f, size_t i)
  : index(i),
    fa(f)
{
}

template <class T>
FileArrayProxy<T>::operator T() const
{
  return fa.readElement(index);
}

template <class T>
FileArrayProxy<T>& FileArrayProxy<T>::operator=(const T& t)
{
  fa.storeElement(index,t);
  return *this;
}

template <class T>
FileArrayProxy<T>& FileArrayProxy<T>::operator=(const FileArrayProxy<T>& p)
{
  fa.storeElement(index,p.fa.readElement(p.index));
  return *this;
}

template <class T>
FileArray<T>::FileArray(const char* name, size_t size)
  : pstream(new fstream(name, ios::in|ios::out)),
    array_size(size)
{
  // what if the file could not be opened?

  // store the size on file.
  (*pstream).write((const char*)&array_size, sizeof(array_size));
  // what if write failed?

  T t;
  storeElement(array_size-1,t);
  // What if this failed?
}

template <class T>
FileArray<T>::FileArray(const char* name)
  : pstream(new fstream(name, ios::in|ios::out)),
    array_size(0)
{
  (*pstream).read((char*)&array_size, sizeof(array_size));
  // what if read fails or array_size == 0?
}

template <class T>
T FileArray<T>::operator[](size_t size) const
{
  // what if size >= array_size?
  return readElement(size);
}

template <class T>
FileArrayProxy<T> FileArray<T>::operator[](size_t size)
{
  // what if size >= array_size?
  return FileArrayProxy<T>(*this , size);
}

template <class T>
size_t FileArray<T>::size() const
{
  return array_size;
}

template <class T>
T FileArray<T>::readElement(size_t index) const
{
  // what if index >= max_elements?
  typedef FileArrayElementAccess<T> traits;
  (*pstream).seekg(sizeof(array_size)+index*traits::size, ios::beg);
  // what if seek fails?

  return traits::readFrom(*pstream);
  // what if read fails?
  // What if too much data is read?
}

template <class T>
void FileArray<T>::storeElement(size_t index, const T& element)
{
  // what if index >= array_size?
  typedef FileArrayElementAccess<T> traits;
  (*pstream).seekp(traits::size*index
		   +sizeof(array_size), ios::beg);
  // what if seek fails?
  traits::writeTo(element,*pstream);
  // what if write failed?
  // what if too much data was written?
}

class FileArrayElementAccess<char>
{
public:
  typedef char type;
  enum { size = sizeof(type) };
  static void writeTo(type c, ostream& os) { os.put(c); };
  static type readFrom(istream& is) { type c; is.get(c); return c; };
};

class FileArrayElementAccess<signed char>
{
public:
  typedef signed char type;
  enum { size = sizeof(type) };
  static void writeTo(type c, ostream& os) { os.put(c); };
  static type readFrom(istream& is) { type c; is.get(c); return c; };
};

class FileArrayElementAccess<unsigned char>
{
public:
  typedef unsigned char type;
  enum { size = sizeof(type) };
  static void writeTo(type c, ostream& os) { os.put(c); };
  static type readFrom(istream& is) { type c; is.get(c); return c; };
};

class FileArrayElementAccess<short>
{
public:
  typedef short type;
  enum { size = sizeof(type) };
  static void writeTo(type v, ostream& os) {os.write((const char*)&v,size);};
  static type readFrom(istream& is){type t;is.read((char*)&t,size);return t;};
};

class FileArrayElementAccess<unsigned short> 
{
public:
  typedef unsigned short type;
  enum { size = sizeof(type) };
  static void writeTo(type v, ostream& os) {os.write((const char*)&v,size);};
  static type readFrom(istream& is){type t;is.read((char*)&t,size);return t;};
};

class FileArrayElementAccess<int> 
{
public:
  typedef int type;
  enum { size = sizeof(type) };
  static void writeTo(type v, ostream& os) {os.write((const char*)&v,size);};
  static type readFrom(istream& is){type t;is.read((char*)&t,size);return t;};
};

class FileArrayElementAccess<unsigned>
{
public:
  typedef unsigned type;
  enum { size = sizeof(type) };
  static void writeTo(type v, ostream& os) {os.write((const char*)&v,size);};
  static type readFrom(istream& is){type t;is.read((char*)&t,size);return t;};
};

class FileArrayElementAccess<long>
{
public:
  typedef long type;
  enum { size = sizeof(type) };
  static void writeTo(type v, ostream& os) {os.write((const char*)&v,size);};
  static type readFrom(istream& is){type t;is.read((char*)&t,size);return t;};
};

class FileArrayElementAccess<unsigned long>
{
public:
  typedef unsigned long type;
  enum { size = sizeof(type) };
  static void writeTo(type v, ostream& os) {os.write((const char*)&v,size);};
  static type readFrom(istream& is){type t;is.read((char*)&t,size); return t;};
};

class FileArrayElementAccess<float>
{
public:
  typedef float type;
  enum { size = sizeof(type) };
  static void writeTo(type v, ostream& os) {os.write((const char*)&v,size);};
  static type readFrom(istream& is){type t;is.read((char*)&t,size);return t;};
};

class FileArrayElementAccess<double>
{
public:
  typedef double type;
  enum { size = sizeof(type) };
  static void writeTo(type v, ostream& os) {os.write((const char*)&v,size);};
  static type readFrom(istream& is){type t;is.read((char*)&t,size);return t;};
};

class FileArrayElementAccess<long double>
{
public:
  typedef long double type;
  enum { size = sizeof(type) };
  static void writeTo(type v, ostream& os) {os.write((const char*)&v,size);};
  static type readFrom(istream& is){type t;is.read((char*)&t,size);return t;};
};
  
#endif // FARRAY_HPP

