#include "farray.hpp"

static const char ArrayBegin[]="ArrayBegin";
static const char ArrayEnd[]="ArrayEnd";

FileArrayBase::temporaryfile::temporaryfile()
  : name(::tempnam(".","array")),
    fs(name,ios::in|ios::out|ios::binary)
{
  // what if tmpnam fails and name is 0
  // what if fs is bad?
}

FileArrayBase::temporaryfile::~temporaryfile()
{
  fs.close();
  ::remove(name);
  // What if remove fails?
  ::free(name);
}

iostream& FileArrayBase::temporaryfile::stream()
{
  return fs;
}

FileArrayBase::FileArrayBase(iostream& io,
			     size_t elements,
			     size_t elem_size)
  throw (FileArrayCreateError,
	 FileArrayDataCorruptionError)
  : stream(io),
    array_size(elements),
    e_size(elem_size)
{
  char pattern[sizeof(ArrayBegin)];
  stream.read(pattern,sizeof(pattern));
  if (stream.eof()) {
    stream.clear(); // clear error state
                    // and initialize.
    
    // begin of array pattern.
    stream.write(ArrayBegin,sizeof(ArrayBegin));

    // must store size of elements
    stream.write((const char*)&elem_size,
		 sizeof(elem_size));

    // and of course the number of elements
    stream.write((const char*)&array_size,
		 sizeof(array_size));

    // Now that we've written the maintenance
    // stuff, we know what the home position is.
    home = stream.tellp();

    // Then we must go the the end and write
    // the end pattern.

    stream.seekp(home+elem_size*array_size);
    stream.write(ArrayEnd,sizeof(ArrayEnd));

    if (!stream) throw FileArrayCreateError();

    // set put and get pointer to past the end pos.
    stream.seekg(stream.tellp());
    
    return;
  }
  initFromFile(pattern); // shared with other
                         // stream constructor

  if (array_size != elements
      || e_size != elem_size) {
    throw FileArrayDataCorruptionError();
  }
  // set put and get pointer to past the end pos.
  stream.seekp(stream.tellg());
}

FileArrayBase::FileArrayBase(iostream& io)
  throw (FileArrayCreateError,
	 FileArrayDataCorruptionError)
  : stream(io)
{
  char pattern[sizeof(ArrayBegin)];
  stream.read(pattern,sizeof(pattern));
  initFromFile(pattern);
  // set put and get pointer to past the end pos.
  stream.seekp(stream.tellg());
}

FileArrayBase::FileArrayBase(size_t elements,
			     size_t elem_size)
  throw (FileArrayCreateError)
  : pfile(new temporaryfile),
    stream(pfile->stream()),
    array_size(elements),
    e_size(elem_size),
    home(stream.tellg())
{
  stream.seekp(home+array_size*e_size);
  char c;
  stream.write(&c,1);
  if (!stream) throw FileArrayCreateError();
  // set put and get pointer to past the end pos.
  stream.seekg(stream.tellp());
}
void FileArrayBase::initFromFile(const char* p)
    throw (FileArrayDataCorruptionError)
{
  // Check if the read pattern is correct
  if (strncmp(p,ArrayBegin,sizeof(ArrayBegin))) {
    throw FileArrayDataCorruptionError();
  }
  // OK, we have a valid array, now let's see if
  // it's of the right kind.
  stream.read((char*)&e_size,sizeof(e_size));
  
  // Get the size of the array. Can't do much with
  // the size here, though.
  stream.read((char*)&array_size,sizeof(array_size));
  
  // Now we're past the header, so we know where the
  // data begins and can set the home position.
  
  home = stream.tellg();
  
  stream.seekg(home+e_size*array_size);
  // Now positioned immediately after the last
  // element.
  char epattern[sizeof(ArrayEnd)];
  stream.read(epattern,sizeof(epattern));
  if (strncmp(epattern,ArrayEnd,sizeof(ArrayEnd))) {
    throw FileArrayDataCorruptionError();
  }
  // Seems like we have a valid array!
}
  
iostream& FileArrayBase::seekg(size_t index) const
  throw (FileArrayBoundsError,
	 FileArrayStreamError)
{
  // what if index is out of bounds?
  if (index > array_size) 
    throw FileArrayBoundsError();
  stream.seekg(home+index*e_size);
  // what if seek failed?
  if (!stream)
    throw FileArrayStreamError();
  return stream;
}

iostream& FileArrayBase::seekp(size_t index) const
  throw (FileArrayBoundsError,
	 FileArrayStreamError)
{
  // what if index is out of bounds?
  if (index > array_size) 
    throw FileArrayBoundsError();
  stream.seekp(home+index*e_size);
  // what if seek failed?
  if (!stream)
    throw FileArrayStreamError();
  return stream;
}

size_t FileArrayBase::size() const throw()
{
  return array_size;
}

size_t FileArrayBase::element_size() const throw()
{
  return e_size;
}

