#ifndef FARRAY_HPP
#define FARRAY_HPP

#include <fstream.h>
#include <stdlib.h> // size_t
#include <string.h> // strncmp
#include <stdio.h> // tempnam

class FileArrayException {};
class FileArrayLogicError : public FileArrayException {};
class FileArrayRuntimeError : public FileArrayException {};
class FileArrayCreateError : public FileArrayRuntimeError {};
class FileArrayStreamError : public FileArrayRuntimeError {};
class FileArrayDataCorruptionError : public FileArrayRuntimeError {};
class FileArrayBoundsError : public FileArrayLogicError {};
class FileArrayElementSizeError : public FileArrayLogicError {};

template <class T> class FileArrayElementAccess {};

template <class T> class FileArray;
template <class T> class FileArrayProxy
{
public:
  operator T() const;
  FileArrayProxy<T>& operator=(const T& t);
  FileArrayProxy<T>& operator=(const FileArrayProxy<T>& p);
private:
  FileArrayProxy(FileArray<T>& f, size_t i);

  operator T&() const;
  T* operator&(void) const;
  size_t index;
  FileArray<T>& fa;

  friend class FileArray<T>;
};

template <class T>
ostream& operator<<(ostream& os, const FileArrayProxy<T>& p)
{
  return os << p.operator T();
}

template <class T>
class ptr
{
public:
  ptr(T* pt=0) : p(pt) {};
  ~ptr() { delete p; };
  T* operator->(void) const { return p; };
  T& operator*(void) const { return *p; };
private:
  ptr(const ptr<T>&);
  ptr<T>& operator=(const ptr<T>&);
  T* p;
};

class FileArrayBase
{
public:
protected:
  FileArrayBase(iostream& io,
		size_t elements,
		size_t elem_size)
    throw (FileArrayCreateError,
	   FileArrayDataCorruptionError);
  FileArrayBase(iostream& io)
    throw (FileArrayCreateError,
	   FileArrayDataCorruptionError);
  FileArrayBase(size_t elements, size_t elem_size)
    throw (FileArrayCreateError);
  iostream& seekp(size_t index) const
    throw (FileArrayBoundsError,
	   FileArrayStreamError);
  iostream& seekg(size_t index) const
    throw (FileArrayBoundsError,
	   FileArrayStreamError);
  
  size_t size() const throw(); // number of elements
  size_t element_size() const throw();
private:
  class temporaryfile
  {
  public:
    temporaryfile();
    ~temporaryfile();
    iostream& stream();
  private:
    char* name;
    fstream fs;
  };
  void initFromFile(const char* p)
    throw (FileArrayDataCorruptionError);
  ptr<temporaryfile> pfile;
  iostream& stream;
  size_t array_size;
  size_t e_size;
  streampos home;
};

template <class T>
class FileArrayIterator;

template <class T>
class FileArray : private FileArrayBase
{
public:
  typedef FileArrayIterator<T> iterator;
  FileArray(iostream& fs, size_t elements); // create one.
  FileArray(iostream& fs); // use existing array
  FileArray(size_t elements); // use temporary file
  ~FileArray(void);
  T operator[](size_t index) const;
  FileArrayProxy<T> operator[](size_t index);
  size_t size() const { return FileArrayBase::size();};
  iterator begin() { return *this; };
  iterator end() { iterator i(*this); return i+=size();};
private:
  FileArray(const FileArray<T>&); // illegal
  FileArray<T>& operator=(const FileArray<T>&); // illegal
  
  void initFromFile(const char*);
  
  T readElement(size_t index) const;
  void storeElement(size_t index, const T& elem);
  
  friend class FileArrayProxy<T>;
};

template <class T>
class FileArrayIterator
{
public:
  FileArrayIterator(FileArray<T>& f);
  T operator*() const;
  FileArrayProxy<T> operator*();
  T operator[](long) const;
  FileArrayProxy<T> operator[](long);
  FileArrayIterator<T>& operator+=(long);
  FileArrayIterator<T>& operator-=(long);
  FileArrayIterator<T>& operator++();
  FileArrayIterator<T> operator++(int);
  FileArrayIterator<T>& operator--();
  FileArrayIterator<T> operator--(int);
  int operator==(const FileArrayIterator<T>&) const;
  int operator!=(const FileArrayIterator<T>&) const;
  int operator<(const FileArrayIterator<T>&)  const;
  int operator<=(const FileArrayIterator<T>&) const;
  int operator>(const FileArrayIterator<T>&)  const;
  int operator>=(const FileArrayIterator<T>&) const;
private:
  FileArray<T>* array;
  long index;
};

template <class T>
FileArrayIterator<T> operator+(const FileArrayIterator<T>&,long);

template <class T>
FileArrayIterator<T> operator+(long, const FileArrayIterator<T>&);

template <class T>
FileArrayIterator<T> operator-(const FileArrayIterator<T>&,long);

template <class T>
FileArrayIterator<T> operator-(long, const FileArrayIterator<T>&);

template <class T>
FileArrayIterator<T> operator+(FileArray<T>&,long);

template <class T>
FileArrayIterator<T> operator+(long,FileArray<T>&);

template <class T>
FileArrayIterator<T>::FileArrayIterator(FileArray<T>& f)
  : array(&f),
    index(0)
{
}

template <class T>
T FileArrayIterator<T>::operator*() const
{
  return (*(const FileArray<T>*)array)[index];
}

template <class T>
FileArrayProxy<T> FileArrayIterator<T>::operator*()
{
  return (*array)[index];
}

template <class T>
T FileArrayIterator<T>::operator[](long l) const
{
  return (*(const FileArray<T>*)array)[index+l];
}

template <class T>
FileArrayProxy<T> FileArrayIterator<T>::operator[](long l)
{
  return (*array)[index+l];
}

template <class T>
FileArrayIterator<T>& FileArrayIterator<T>::operator+=(long l)
{
  index+=l;
  return *this;
}

template <class T>
FileArrayIterator<T>& FileArrayIterator<T>::operator-=(long l)
{
  index-=l;
  return *this;
}

template <class T>
FileArrayIterator<T>& FileArrayIterator<T>::operator++()
{
  ++index;
  return *this;
}

template <class T>
FileArrayIterator<T> FileArrayIterator<T>::operator++(int)
{
  FileArrayIterator<T> t(*this);
  ++*this;
  return t;
}

template <class T>
FileArrayIterator<T>& FileArrayIterator<T>::operator--()
{
  --index;
  return *this;
}

template <class T>
FileArrayIterator<T> FileArrayIterator<T>::operator--(int)
{
  FileArrayIterator<T> t(*this);
  --*this;
  return t;
}

template <class T>
FileArrayIterator<T> operator+(const FileArrayIterator<T>& i,long l)
{
  FileArrayIterator<T> r(i);
  return r+=l;
}

template <class T>
FileArrayIterator<T> operator+(long l, const FileArrayIterator<T>& i)
{
  FileArrayIterator<T> r(i);
  return r+=l;
}

template <class T>
FileArrayIterator<T> operator-(const FileArrayIterator<T>& i,long l)
{
  FileArrayIterator<T> r(i);
  return r-=l;
}

template <class T>
FileArrayIterator<T> operator-(long l, const FileArrayIterator<T>& i)
{
  FileArrayIterator<T> r(i);
  return r-=l;
}

template <class T>
FileArrayIterator<T> operator+(FileArray<T>& a,long l)
{
  FileArrayIterator<T> i(a);
  return i+=l;
}

template <class T>
FileArrayIterator<T> operator+(long l,FileArray<T>& a)
{
  FileArrayIterator<T> i(a);
  return i+=l;
}

template <class T>
int FileArrayIterator<T>::operator==(const FileArrayIterator<T>& rhs) const
{
  return array == rhs.array && index == rhs.index;
}

template <class T>
int FileArrayIterator<T>::operator!=(const FileArrayIterator<T>& rhs) const
{
  return !(*this == rhs);
}

template <class T>
int FileArrayIterator<T>::operator<(const FileArrayIterator<T>& rhs) const
{
  return array == rhs.array && index < rhs.index;
}

template <class T>
int FileArrayIterator<T>::operator<=(const FileArrayIterator<T>& rhs) const
{
  return array==rhs.array && index <= rhs.index;
}

template <class T>
int FileArrayIterator<T>::operator>(const FileArrayIterator<T>& rhs) const
{
  return array == rhs.array && index > rhs.index;
}

template <class T>
int FileArrayIterator<T>::operator>=(const FileArrayIterator<T>& rhs) const
{
  return array==rhs.array && index >= rhs.index;
}


template <class T>
FileArrayProxy<T>::FileArrayProxy(FileArray<T>& f, size_t i)
  : index(i),
    fa(f)
{
}

template <class T>
FileArrayProxy<T>::operator T() const
{
  return fa.readElement(index);
}

template <class T>
FileArrayProxy<T>& FileArrayProxy<T>::operator=(const T& t)
{
  fa.storeElement(index,t);
  return *this;
}

template <class T>
FileArrayProxy<T>& FileArrayProxy<T>::operator=(const FileArrayProxy<T>& p)
{
  fa.storeElement(index,p.fa.readElement(p.index));
  return *this;
}

template <class T>
FileArray<T>::~FileArray()
{
}

template <class T>
FileArray<T>::FileArray(iostream& io, size_t elements)
  : FileArrayBase(io,
		  elements,
		  FileArrayElementAccess<T>::size)
{
}

template <class T>
FileArray<T>::FileArray(iostream& io)
  : FileArrayBase(io)
{
  // what if the element size is wrong?
  // Handled by base class.
}

template <class T>
FileArray<T>::FileArray(size_t elements)
  : FileArrayBase(elements,
		  FileArrayElementAccess<T>::size)
{
}

template <class T>
T FileArray<T>::operator[](size_t size) const
{
  // what if size >= max_size?
  // readElement takes care of that.
  return readElement(size);
}

template <class T>
FileArrayProxy<T> FileArray<T>::operator[](size_t size)
{
  // what if size >= max_size?
  // The proxy does (indirectly) take care of that.
  return FileArrayProxy<T>(*this , size);
}

template <class T>
T FileArray<T>::readElement(size_t index) const
{
  // what if index >= max_elements?
  // seekg will throw.
  iostream& s = seekg(index); // parent seekg
  // what if seek fails?
  if (!s) throw FileArrayStreamError();
  streampos p1 = s.tellg();
  T t = FileArrayElementAccess<T>::readFrom(s);
  // what if read fails?
  if (!s) throw FileArrayStreamError();
  streampos p2 = s.tellg();
  // What if too much data is read?
  if (p2-p1 > element_size())
    throw FileArrayElementSizeError();
  return t;
}

template <class T>
void FileArray<T>::storeElement(size_t index, const T& element)
{
  // what if index >= array_size?
  // seekp will throw.
  iostream& s = seekp(index); // parent seekp
  // what if seek fails?
  if (!s) throw FileArrayStreamError();
  streampos p1=s.tellp();
  FileArrayElementAccess<T>::writeTo(element,s);
  // what if write failed?
  if (!s) throw FileArrayStreamError();
  streampos p2=s.tellp();
  // what if too much data was written?
  if (p2-p1 > element_size())
    throw FileArrayElementSizeError();
}

class FileArrayElementAccess<char>
{
public:
  typedef char type;
  enum { size = sizeof(type) };
  static void writeTo(type c, ostream& os) { os.put(c); };
  static type readFrom(istream& is) { type c; is.get(c); return c; };
};

class FileArrayElementAccess<signed char>
{
public:
  typedef signed char type;
  enum { size = sizeof(type) };
  static void writeTo(type c, ostream& os) { os.put(c); };
  static type readFrom(istream& is) { type c; is.get(c); return c; };
};

class FileArrayElementAccess<unsigned char>
{
public:
  typedef unsigned char type;
  enum { size = sizeof(type) };
  static void writeTo(type c, ostream& os) { os.put(c); };
  static type readFrom(istream& is) { type c; is.get(c); return c; };
};

class FileArrayElementAccess<short>
{
public:
  typedef short type;
  enum { size = sizeof(type) };
  static void writeTo(type v, ostream& os) {os.write((const char*)&v,size);};
  static type readFrom(istream& is){type t;is.read((char*)&t,size);return t;};
};

class FileArrayElementAccess<unsigned short> 
{
public:
  typedef unsigned short type;
  enum { size = sizeof(type) };
  static void writeTo(type v, ostream& os) {os.write((const char*)&v,size);};
  static type readFrom(istream& is){type t;is.read((char*)&t,size);return t;};
};

class FileArrayElementAccess<int> 
{
public:
  typedef int type;
  enum { size = sizeof(type) };
  static void writeTo(type v, ostream& os) {os.write((const char*)&v,size);};
  static type readFrom(istream& is){type t;is.read((char*)&t,size);return t;};
};

class FileArrayElementAccess<unsigned>
{
public:
  typedef unsigned type;
  enum { size = sizeof(type) };
  static void writeTo(type v, ostream& os) {os.write((const char*)&v,size);};
  static type readFrom(istream& is){type t;is.read((char*)&t,size);return t;};
};

class FileArrayElementAccess<long>
{
public:
  typedef long type;
  enum { size = sizeof(type) };
  static void writeTo(type v, ostream& os) {os.write((const char*)&v,size);};
  static type readFrom(istream& is){type t;is.read((char*)&t,size);return t;};
};

class FileArrayElementAccess<unsigned long>
{
public:
  typedef unsigned long type;
  enum { size = sizeof(type) };
  static void writeTo(type v, ostream& os) {os.write((const char*)&v,size);};
  static type readFrom(istream& is){type t;is.read((char*)&t,size); return t;};
};

class FileArrayElementAccess<float>
{
public:
  typedef float type;
  enum { size = sizeof(type) };
  static void writeTo(type v, ostream& os) {os.write((const char*)&v,size);};
  static type readFrom(istream& is){type t;is.read((char*)&t,size);return t;};
};

class FileArrayElementAccess<double>
{
public:
  typedef double type;
  enum { size = sizeof(type) };
  static void writeTo(type v, ostream& os) {os.write((const char*)&v,size);};
  static type readFrom(istream& is){type t;is.read((char*)&t,size);return t;};
};

class FileArrayElementAccess<long double>
{
public:
  typedef long double type;
  enum { size = sizeof(type) };
  static void writeTo(type v, ostream& os) {os.write((const char*)&v,size);};
  static type readFrom(istream& is){type t;is.read((char*)&t,size);return t;};
};
  
#endif // FARRAY_HPP

