#ifndef FARRAY_HPP
#define FARRAY_HPP

#include <fstream.h>
#include <stdlib.h> // size_t
#include <string.h> // strncmp
#include <stdio.h> // tempnam

template <class T> class FileArrayElementAccess {};

template <class T> class FileArray;
template <class T> class FileArrayProxy
{
public:
  operator T() const;
  FileArrayProxy<T>& operator=(const T& t);
  FileArrayProxy<T>& operator=(const FileArrayProxy<T>& p);
private:
  FileArrayProxy(FileArray<T>& f, size_t i);

  operator T&() const;
  T* operator&(void) const;
  size_t index;
  FileArray<T>& fa;

  friend class FileArray<T>;
};

template <class T>
ostream& operator<<(ostream& os, const FileArrayProxy<T>& p)
{
  return os << p.operator T();
}

template <class T>
class ptr
{
public:
  ptr(T* pt=0) : p(pt) {};
  ~ptr() { delete p; };
  T* operator->(void) const { return p; };
  T& operator*(void) const { return *p; };
private:
  ptr(const ptr<T>&);
  ptr<T>& operator=(const ptr<T>&);
  T* p;
};

class FileArrayBase
{
public:
protected:
  FileArrayBase(iostream& io,
		size_t elements,
		size_t elem_size);
  FileArrayBase(iostream& io);
  FileArrayBase(size_t elements, size_t elem_size);
  iostream& seekp(size_t index) const;
  iostream& seekg(size_t index) const;
  size_t size() const; // number of elements
  size_t element_size() const;
private:
  class temporaryfile
  {
  public:
    temporaryfile();
    ~temporaryfile();
    iostream& stream();
  private:
    char* name;
    fstream fs;
  };
  void initFromFile(const char* p);
  ptr<temporaryfile> pfile;
  iostream& stream;
  size_t array_size;
  size_t e_size;
  streampos home;
};

template <class T>
class FileArray : private FileArrayBase
{
public:
  FileArray(iostream& fs, size_t elements); // create one.
  FileArray(iostream& fs); // use existing array
  FileArray(size_t elements); // use temporary file
  ~FileArray(void);
  T operator[](size_t index) const;
  FileArrayProxy<T> operator[](size_t index);
  size_t size() const { return FileArrayBase::size();};
private:
  FileArray(const FileArray<T>&); // illegal
  FileArray<T>& operator=(const FileArray<T>&); // illegal
  
  void initFromFile(const char*);
  
  T readElement(size_t index) const;
  void storeElement(size_t index, const T& elem);
  
  friend class FileArrayProxy<T>;
};


template <class T>
FileArrayProxy<T>::FileArrayProxy(FileArray<T>& f, size_t i)
  : index(i),
    fa(f)
{
}

template <class T>
FileArrayProxy<T>::operator T() const
{
  return fa.readElement(index);
}

template <class T>
FileArrayProxy<T>& FileArrayProxy<T>::operator=(const T& t)
{
  fa.storeElement(index,t);
  return *this;
}

template <class T>
FileArrayProxy<T>& FileArrayProxy<T>::operator=(const FileArrayProxy<T>& p)
{
  fa.storeElement(index,p.fa.readElement(p.index));
  return *this;
}

template <class T>
FileArray<T>::~FileArray()
{
}

template <class T>
FileArray<T>::FileArray(iostream& io, size_t elements)
  : FileArrayBase(io,
		  elements,
		  FileArrayElementAccess<T>::size)
{
}

template <class T>
FileArray<T>::FileArray(iostream& io)
  : FileArrayBase(io)
{
  // what if the element size is wrong?
}

template <class T>
FileArray<T>::FileArray(size_t elements)
  : FileArrayBase(elements,
		  FileArrayElementAccess<T>::size)
{
}

template <class T>
T FileArray<T>::operator[](size_t index) const
{
  // what if index >= size()?
  return readElement(index);
}

template <class T>
FileArrayProxy<T> FileArray<T>::operator[](size_t index)
{
  // what if index >= size()?
  return FileArrayProxy<T>(*this, index);
}

template <class T>
T FileArray<T>::readElement(size_t index) const
{
  // what if index >= size()?
  iostream& s = seekg(index); // parent seekg
  // what if seek fails?
  return FileArrayElementAccess<T>::readFrom(s);
  // what if read fails?
  // What if too much data is read?
}

template <class T>
void FileArray<T>::storeElement(size_t index, const T& element)
{
  // what if index >= size()?
  iostream& s = seekp(index); // parent seekp
  // what if seek fails?
  FileArrayElementAccess<T>::writeTo(element,s);
  // what if wite failed?
  // what if too much data was written?
}

class FileArrayElementAccess<char>
{
public:
  typedef char type;
  enum { size = sizeof(type) };
  static void writeTo(type c, ostream& os) { os.put(c); };
  static type readFrom(istream& is) { type c; is.get(c); return c; };
};

class FileArrayElementAccess<signed char>
{
public:
  typedef signed char type;
  enum { size = sizeof(type) };
  static void writeTo(type c, ostream& os) { os.put(c); };
  static type readFrom(istream& is) { type c; is.get(c); return c; };
};

class FileArrayElementAccess<unsigned char>
{
public:
  typedef unsigned char type;
  enum { size = sizeof(type) };
  static void writeTo(type c, ostream& os) { os.put(c); };
  static type readFrom(istream& is) { type c; is.get(c); return c; };
};

class FileArrayElementAccess<short>
{
public:
  typedef short type;
  enum { size = sizeof(type) };
  static void writeTo(type v, ostream& os) {os.write((const char*)&v,size);};
  static type readFrom(istream& is){type t;is.read((char*)&t,size);return t;};
};

class FileArrayElementAccess<unsigned short> 
{
public:
  typedef unsigned short type;
  enum { size = sizeof(type) };
  static void writeTo(type v, ostream& os) {os.write((const char*)&v,size);};
  static type readFrom(istream& is){type t;is.read((char*)&t,size);return t;};
};

class FileArrayElementAccess<int> 
{
public:
  typedef int type;
  enum { size = sizeof(type) };
  static void writeTo(type v, ostream& os) {os.write((const char*)&v,size);};
  static type readFrom(istream& is){type t;is.read((char*)&t,size);return t;};
};

class FileArrayElementAccess<unsigned>
{
public:
  typedef unsigned type;
  enum { size = sizeof(type) };
  static void writeTo(type v, ostream& os) {os.write((const char*)&v,size);};
  static type readFrom(istream& is){type t;is.read((char*)&t,size);return t;};
};

class FileArrayElementAccess<long>
{
public:
  typedef long type;
  enum { size = sizeof(type) };
  static void writeTo(type v, ostream& os) {os.write((const char*)&v,size);};
  static type readFrom(istream& is){type t;is.read((char*)&t,size);return t;};
};

class FileArrayElementAccess<unsigned long>
{
public:
  typedef unsigned long type;
  enum { size = sizeof(type) };
  static void writeTo(type v, ostream& os) {os.write((const char*)&v,size);};
  static type readFrom(istream& is){type t;is.read((char*)&t,size); return t;};
};

class FileArrayElementAccess<float>
{
public:
  typedef float type;
  enum { size = sizeof(type) };
  static void writeTo(type v, ostream& os) {os.write((const char*)&v,size);};
  static type readFrom(istream& is){type t;is.read((char*)&t,size);return t;};
};

class FileArrayElementAccess<double>
{
public:
  typedef double type;
  enum { size = sizeof(type) };
  static void writeTo(type v, ostream& os) {os.write((const char*)&v,size);};
  static type readFrom(istream& is){type t;is.read((char*)&t,size);return t;};
};

class FileArrayElementAccess<long double>
{
public:
  typedef long double type;
  enum { size = sizeof(type) };
  static void writeTo(type v, ostream& os) {os.write((const char*)&v,size);};
  static type readFrom(istream& is){type t;is.read((char*)&t,size);return t;};
};
  
#endif // FARRAY_HPP

